export declare type Predicate<T> = {
    (item: T): boolean;
    description?: string;
};
export declare type Action<T> = (item: T) => void;
export interface Node<T> {
    /**
     * Creates a new node and add a predicated link to it.
     * The path is taken if *any* predicate matches.
     * Returns the new node.
     */
    addPath(...predicates: Predicate<T>[]): Node<T>;
    /**
     * Adds a predicated link to the specified node.
     */
    addPathTo(predicate: Predicate<T>, target: Node<T>): void;
    /**
     * Adds an action to occur on this node if none of its links are followed.
     */
    addTerminalAction(action: Action<T>): void;
    /**
     * Adds an action to always occur on this node if it is traversed.
     */
    addAlwaysAction(action: Action<T>): void;
    /**
     * Creates a node that is traversed to if no other links are traversed.
     */
    otherwise(): Node<T>;
    /**
     * Process this item by following any matched links and executing any actions.
     */
    process(item: T, catchHandler?: any): void;
    /**
     * Adds a description to this node.
     * Returns the same node.
     */
    describe(name: string): Node<T>;
    /**
     * Attaches a handler to be invoked if a node fails to follow a defined traversal of the graph.
     */
    catch(handler: (erroringItem: T, error: TreeageError) => void): void;
    /**
     * Returns true if this node has any paths which would be traversed for this item.
     */
    groupingPredicate(item: T): boolean;
    readonly description: string;
    readonly hitCount: number;
    readonly terminalHitCount: number;
}
export declare type ImplementedNode<T> = Node<T> & {
    paths: ReadonlyArray<[readonly Predicate<T>[], Node<T>]>;
    terminalActions: ReadonlyArray<Action<T>>;
    actions: ReadonlyArray<Action<T>>;
    otherwiseNode: Node<T> | undefined;
};
export declare type TreeageError = MultiplePathsMatchedError;
export declare type MultiplePathsMatchedError = {
    type: "multi-path";
    predicates: ReadonlyArray<Predicate<any>>;
};
export interface NodeOptions {
    pathMode: "first" | "single" | "multiple";
}
export declare function createGroupNode<T>(): {
    node: Node<T>;
    predicate: (item: T) => void;
    addGroupChild: (predicate: Predicate<T>) => void;
};
export declare function create<T>(nodeOptions?: NodeOptions): Node<T>;
export declare function or<T>(...preds: Predicate<T>[]): Predicate<T>;
export declare function visualizeNodeTree(root: Node<any>, includeLineLabels?: boolean): string;
