"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
function assertNever(x) {
    throw new Error(`Impossible value ${x} observed`);
}
function createGroupNode() {
    const node = create();
    const preds = [];
    function addGroupChild(predicate) {
        preds.push(predicate);
    }
    function predicate(item) {
        throw new Error("Not implemented");
    }
    return {
        node,
        predicate,
        addGroupChild
    };
}
exports.createGroupNode = createGroupNode;
function create(nodeOptions) {
    const options = Object.assign({ pathMode: "single" }, nodeOptions);
    const paths = [];
    const terminalActions = [];
    const actions = [];
    let catchHandler = undefined;
    let description = "(unnamed)";
    let otherwiseNode = undefined;
    let hitCount = 0;
    let terminalHitCount = 0;
    function describe(name) {
        description = name;
        return impl;
    }
    function process(item, parentCatcher = undefined) {
        hitCount++;
        let alreadyMatched = false;
        const matchedPaths = [];
        const matchedPreds = [];
        for (const path of paths) {
            for (const pred of path[0]) {
                if (pred(item)) {
                    matchedPreds.push(pred);
                    matchedPaths.push(path);
                    break;
                }
            }
        }
        if (matchedPaths.length > 0) {
            switch (options.pathMode) {
                case "first":
                    matchedPaths[0][1].process(item, catchHandler || parentCatcher);
                    break;
                case "multiple":
                    for (const p of matchedPaths) {
                        p[1].process(item, catchHandler || parentCatcher);
                    }
                    break;
                case "single":
                    if (matchedPaths.length !== 1) {
                        const err = {
                            type: "multi-path",
                            predicates: matchedPreds
                        };
                        const catchFunc = catchHandler || parentCatcher;
                        if (catchFunc) {
                            catchFunc(item, err);
                        }
                        else {
                            throw err;
                        }
                    }
                    else {
                        matchedPaths[0][1].process(item, catchHandler || parentCatcher);
                    }
                    break;
                default:
                    assertNever(options.pathMode);
                    break;
            }
        }
        else {
            if (otherwiseNode) {
                otherwiseNode.process(item, catchHandler || parentCatcher);
            }
            else {
                terminalHitCount++;
                for (const term of terminalActions) {
                    term(item);
                }
            }
        }
        for (const always of actions) {
            always(item);
        }
    }
    function addPath(...predicates) {
        const target = create(nodeOptions);
        if (predicates[0].description) {
            target.describe(predicates[0].description);
        }
        paths.push([predicates, target]);
        return target;
    }
    function addPathTo(predicate, target) {
        paths.push([[predicate], target]);
    }
    function addTerminalAction(action) {
        terminalActions.push(action);
    }
    function addAlwaysAction(action) {
        actions.push(action);
    }
    function otherwise() {
        if (otherwiseNode)
            throw new Error("Cannot call 'otherwise' twice on this node");
        return otherwiseNode = create(nodeOptions);
    }
    function catchImpl(handler) {
        catchHandler = handler;
    }
    function groupingPredicate(item) {
        return paths.some(p => p[0].some(p => p(item)));
    }
    Object.defineProperties(groupingPredicate, {
        description: {
            get() {
                return paths.map(p => p[0][0].description || "??").join(" or ");
            }
        }
    });
    const impl = {
        addPath,
        addPathTo,
        addTerminalAction,
        addAlwaysAction,
        otherwise,
        describe,
        paths,
        actions,
        terminalActions,
        process,
        groupingPredicate,
        catch: catchImpl,
        get description() {
            return description;
        },
        get otherwiseNode() {
            return otherwiseNode;
        },
        get hitCount() {
            return hitCount;
        },
        get terminalHitCount() {
            return terminalHitCount;
        }
    };
    return impl;
}
exports.create = create;
function or(...preds) {
    function result(item) {
        return preds.some(p => p(item));
    }
    result.description = preds.map(p => p.description).join(" or ");
    return result;
}
exports.or = or;
function visualizeNodeTree(root, includeLineLabels = false) {
    const seenNodes = [];
    const lines = [];
    lines.push(`digraph treeage {`);
    lines.push(`    graph [splines = ortho];`);
    lines.push(`    node [shape = record];`);
    recordNode(root);
    lines.push(`}`);
    return lines.join("\r\n");
    function recordNode(node) {
        const existing = seenNodes.indexOf(node);
        if (existing >= 0) {
            return `node_${existing}`;
        }
        const nodeName = `node_${seenNodes.push(node) - 1}`;
        lines.push(`    ${nodeName} [label="${node.description} (${node.hitCount})"];`);
        for (const path of node.paths) {
            const target = path[1];
            const targetName = recordNode(target);
            const pred = path[0];
            if (includeLineLabels) {
                lines.push(`    ${nodeName} -> ${targetName} [label="${pred.description || ""}"];`);
            }
            else {
                lines.push(`    ${nodeName} -> ${targetName};`);
            }
        }
        if (node.otherwiseNode) {
            const targetName = recordNode(node.otherwiseNode);
            if (includeLineLabels) {
                lines.push(`    ${nodeName} -> ${targetName} [label="else"];`);
            }
            else {
                lines.push(`    ${nodeName} -> ${targetName};`);
            }
        }
        return nodeName;
    }
}
exports.visualizeNodeTree = visualizeNodeTree;
function getAllParentedNodes(root) {
    const results = [];
    recur(root, "");
    return results;
    function recur(node, prefix) {
        results.push([node, prefix + node.description]);
        for (const p of node.paths) {
            recur(p[1], prefix + node.description + ".");
        }
    }
}
